#include "dp_finder.h"
#include "bpr_finder.h"
#include "bpm_finder.h"
#include "list.h"
#include "common.h"
#include <stdio.h>
#include <string.h>
#include <time.h>

/*
 * Osumafunktiot, jotka lisäävät löydetyn esiintymän listaan.
 */
int dp_count;
void dp_hit(int i) {
	++dp_count;
}

int bpm_count;
void bpm_hit(int i) {
	++bpm_count;
}

int bpr_count;
void bpr_hit(int i) {
	++bpr_count;
}


/*
 * Lukee tiedoston sisälmykset muistiin.
 */
char* read_file(size_t* length, FILE* file) {
	size_t size = 4096;
	char* buffer = malloc(size);
	*length = fread(buffer, 1, size, file);

	while (*length == size) {
		size *= 2;
		buffer = realloc(buffer, size);
		*length += fread(buffer + *length, 1, size / 2, file);
	}
	
	buffer[*length] = '\0';
	
	return buffer;
}


/*
 * Arpoo annetusta tekstistä halutun pituisen satunnaisen osajonon.
 */
char const* random_substring(char const* text, int n, int m) {
	return text + rand() % (n - m + 1);
}


/*
 * Pääohjelma.
 */
int main(int argc, char* args[]) {
	char* text;

	FILE* text_file = stdin;
	int sample_count = 1;
	char* result_prefix = NULL;
	int alength;
	
	// Parsitaan parametrit.
	switch (argc) {
		case 5:
			sample_count = atoi(args[4]);
		case 4:
			text_file = fopen(args[1], "r");
			if (text_file == NULL) {
				fprintf(stderr, "Unabled to open file '%s'\n", args[1]);
				return 1;
			}
			result_prefix = args[2];
			alength = atoi(args[3]);
			break;
		default:
			fprintf(stderr, "usage: %s <test data file> <result prefix> <alphabet length> [<sample count>]\n", args[0]);
			return 1;
	}
	
	// Luetaan tiedosto.
	size_t t_length;
	text = read_file(&t_length, text_file);
	
	// Supistetaan aakkoston koko halutuksi.
	for (int i = 0; i < t_length; ++i)
		((unsigned char*) text)[i] &= (unsigned char) ((1 << alength) - 1);

	//fprint_hex_array(stderr, text, 50);
	
	// Suoritetaan testaus:
	
	// Käydään eri hahmon pituudet läpi. (1, 3, 7, 15, ...)
	char const* pattern;
	int pattern_length_count = 5;
	for (int i = 1; i <= pattern_length_count; ++i) {
		int m = (2 << (i-1)) - 1;
		
		char buffer[128];
		snprintf(buffer, 128, "results/%s_a%d_%02d_dp.txt", result_prefix, alength, m);
		FILE * dp_results = fopen(buffer, "w");
		
		snprintf(buffer, 128, "results/%s_a%d_%02d_bpm.txt", result_prefix, alength, m);
		FILE * bpm_results = fopen(buffer, "w");
		
		snprintf(buffer, 128, "results/%s_a%d_%02d_bpr.txt", result_prefix, alength, m);
		FILE * bpr_results = fopen(buffer, "w");
		
		// Käydään läpi sample_count satunnaista hahmoa.
		for (int p = 1; p <= sample_count; ++p) {
			pattern = random_substring(text, t_length, m);
			
			// Testataan erilaisilla maksimivirhemäärillä.
			for (int k = 0; k <= m; ++k) {
				printf("pattern length: %d/%d, sample: %d/%d, max errors: %d/%d\n",
						i, pattern_length_count, p, sample_count, k, m);
				
				// Rajoitettu dynaaminen ohjelmointi.
				reset_clock();							// Aloitetaan ajanotto.
				dp_count = 0;
				dp_find(text, t_length, pattern, m, k, dp_hit);
				double dp_time_used = time_elapsed();	// lopetetaan ajanotto.
				fprintf(dp_results, "%10f", dp_time_used);
				
				
				// Myersin bittirinnakkainen algoritmi.
				reset_clock();							// Aloitetaan ajanotto.
				bpm_count = 0;
				bpm_find(text, t_length, pattern, m, k, bpm_hit);
				double bpm_time_used = time_elapsed();	// lopetetaan ajanotto.
				fprintf(bpm_results, "%10f", bpm_time_used);
				
				
				// Bittirinnakkainen NFA-simulointi.
				reset_clock();							// Aloitetaan ajanotto.
				bpr_count = 0;
				bpr_find(text, t_length, pattern, m, k, bpr_hit);
				double bpr_time_used = time_elapsed();	// lopetetaan ajanotto.
				fprintf(bpr_results, "%10f", bpr_time_used);
				
				
				// Tarkastetaan täsmäävätkö algoritmien tulokset.
				if (dp_count != bpr_count || dp_count != bpm_count) {
					fprintf(stderr, "Result counts of all algorithms don't match (p=%d, k=%d, m=%d).\n", p, k, m);
					fprintf(stderr, "dp: %d, bpm: %d, dpr: %d\n", dp_count, bpm_count, bpr_count);
					fprintf(stderr, "Pattern: ");
					fprint_hex_array(stderr, pattern, m);
				}
			}
			fprintf(dp_results, "\n");
			fprintf(bpm_results, "\n");
			fprintf(bpr_results, "\n");
		}
		
		fclose(dp_results);
		fclose(bpm_results);
		fclose(bpr_results);
	}
	
	free(text);
	
	return 0;
}

