#include "bpm_finder.h"
#include "common.h"
#include <stdio.h>

/*
 * Myersin bittirinnakkainen algoritmi likimääräiseen hahmon etsintään.
 * Params:
 *		text = Teksti, josta hahmoa etsitään.
 *		t_length = Tekstin pituus.
 *		pattern = Haettava hahmo.
 *		p_length = Hahmon pituus.
 *		max_errors = Suurin sallittu virheiden määrä hahmon esiintymässä.
 *		hit = Funktio, jota kutsutaan, kun löydetään esiintymä.
 */
void bpm_find(const char* text, size_t t_length, const char* pattern, size_t p_length, size_t max_errors, void (*hit)(int)) {
	// Alustetaan aputaulukko.
	BitRow positions[CHAR_SIZE];
	for (int c = 0; c < CHAR_SIZE; ++c)
		positions[c] = 0;
	for (size_t i = 0; i < p_length; ++i)
		positions[(unsigned char)pattern[i]] |= 1ll << i;
	
	// Alustetaan alkutilanne.
	BitRow p_v = (1ll << p_length) - 1;
	BitRow m_v = 0;
	BitRow s_c = p_length;
	
	// Suoritetaan etsintä.
	for (int j = 0; j < t_length; ++j) {
		BitRow eq = positions[(unsigned char)text[j]];
		BitRow x_h = (((eq & p_v) + p_v) ^ p_v) | eq;
		BitRow p_h = m_v | ~(x_h | p_v);
		BitRow m_h = p_v & x_h;
		BitRow x_v = eq | m_v;

		p_v = (m_h << 1) | ~(x_v | (p_h << 1));
		m_v = (p_h << 1) & x_v;
		s_c = s_c + ((p_h >> (p_length - 1)) & 1) - ((m_h >> (p_length - 1)) & 1);
		
		// Tarkistetaan, löytyikö esiintymä.
		if (s_c <= max_errors)
			(*hit)(j);
	}
}
