#include "bpr_finder.h"
#include "common.h"
#include <stdio.h>

/*
 * Wu-Manberin bittirinnakkainen algoritmi likimääräiseen hahmon etsintään.
 * Params:
 *		text = Teksti, josta hahmoa etsitään.
 *		t_length = Tekstin pituus.
 *		pattern = Haettava hahmo.
 *		p_length = Hahmon pituus.
 *		max_errors = Suurin sallittu virheiden määrä hahmon esiintymässä.
 *		hit = Funktio, jota kutsutaan, kun löydetään esiintymä.
 */
void bpr_find(const char* text, size_t t_length, const char* pattern, size_t p_length, size_t max_errors, void (*hit)(int)) {
	// Rajoitetaan suurin sallittu virheiden määrä hahmon pituuteen.
	max_errors = max_errors > p_length ? p_length : max_errors;
	
	// Alustetaan aputaulukko.
	BitRow positions[CHAR_SIZE];
	for (int c = 0; c < CHAR_SIZE; ++c)
		positions[c] = 0;
	for (size_t i = 0; i < p_length; ++i)
		positions[(unsigned char)pattern[i]] |= 1ll << (i + 1);

	// Alustetaan automaatti.
	BitRow rows[max_errors + 1];
	for (size_t y = 0; y <= max_errors; ++y)
		rows[y] = 1ll << y;
		
	// Simuloidaan automaattia tekstillä merkki kerrallaan.
	for (size_t i = 0; i < t_length; ++i) {
		// Päivitetään ensimmäinen rivi.
		BitRow old_row = rows[0];
		BitRow new_row = (((old_row << 1) | 1) & positions[(unsigned char)text[i]]) | 1;
		rows[0] = new_row;
		
		// Päivitetään loput rivit.
		for (int y = 1; y <= max_errors; ++y) {
			new_row = ((rows[y] << 1) & positions[(unsigned char)text[i]]) |
					old_row | ((old_row | new_row) << 1);
			old_row = rows[y];
			rows[y] = new_row;
		}
		
		// Tarkistetaan, löytyikö esiintymä.
		if (new_row & (1ll << p_length))
			(*hit)(i);
	}
}
