#include "dp_finder.h"
#include <stdio.h>

inline int min3(size_t a, size_t b, size_t c) {
	if (a < b)
		return a < c ? a : c;
	else
		return b < c ? b : c;
}

inline int min2(size_t a, size_t b) {
	return a < b ? a : b;
}

/*
 * Ukkosen rajoitettu dynaaminen algoritmi likimääräiseen hahmon etsintään.
 * Params:
 *		text = Teksti, josta hahmoa etsitään.
 *		t_length = Tekstin pituus.
 *		pattern = Haettava hahmo.
 *		p_length = Hahmon pituus.
 *		max_errors = Suurin sallittu virheiden määrä hahmon esiintymässä.
 *		hit = Funktio, jota kutsutaan, kun löydetään esiintymä.
 */
void dp_find(const char* text, size_t t_length, const char* pattern, size_t p_length, size_t max_errors, void (*hit)(int)) {
	int top = min2(max_errors + 1, p_length);
	
	// Alustetaan apurakenne.
	size_t temp_col[p_length + 1];
	for (int i = 0; i <= top; ++i)
		temp_col[i] = i;
	
	// Suoritetaan etsintä.
	for (int col = 0; col < t_length; ++col) {
		int old = 0;	// p_{i-1,j-1}
		for (int row = 1; row <= top; ++row) {
			int next_old = temp_col[row];
			temp_col[row] = min3(
					temp_col[row] + 1,
					temp_col[row - 1] + 1,
					old + (pattern[row - 1] == text[col] ? 0 : 1));
			old = next_old;
		}
		while (temp_col[top] > max_errors)
			--top;
		
		// Tarkistetaan, löytyikö esiintymä.
		if (top == p_length)
			(*hit)(col);
		else {
			++top;
			temp_col[top] = max_errors;
		}
	}	
}
