#include "dp_finder.h"
#include "bpr_finder.h"
#include "bpm_finder.h"
#include "list.h"
#include "common.h"
#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include <ctype.h>

int count = 0;

/*
 * Laskee listan alkion.
 */
void count_node(int num) {
	++count;
}

/*
 * Tulostaa int-muuttujan.
 */
void print_int(int num) {
	printf("%d, ", num);
}


/*
 * Osumafunktiot, jotka lisäävät löydetyn esiintymän listaan.
 */
IntList * dp_results;
void dp_hit(int i) {
	add_to_int_list(dp_results, i);
}

IntList * bpm_results;
void bpm_hit(int i) {
	add_to_int_list(bpm_results, i);
}

IntList * bpr_results;
void bpr_hit(int i) {
	add_to_int_list(bpr_results, i);
}


/*
 * Lukee tiedoston sisälmykset muistiin.
 */
char* read_file(size_t* length, FILE* file) {
	size_t size = 4096;
	char* buffer = malloc(size);
	*length = fread(buffer, 1, size, file);

	while (*length == size) {
		size *= 2;
		buffer = realloc(buffer, size);
		*length += fread(buffer + *length, 1, size / 2, file);
	}
	
	buffer[*length] = '\0';
	
	return buffer;
}

/*
 * Pääohjelma.
 */
int main(int argc, char* args[]) {
	int k = 0;
	int occ_count = 10;
	char* text;
	char* pattern;
	FILE* text_file = stdin;
	
	// Parsitaan parametrit.
	switch (argc) {
		case 5:
			occ_count = atoi(args[4]);
		case 4:
			if (strcmp(args[3], "-") != 0) {
				text_file = fopen(args[3], "r");
				if (text_file == NULL) {
					fprintf(stderr, "Unabled to open file '%s'\n", args[3]);
					return 1;
				}
			}
		case 3:
			k = atoi(args[2]);
			pattern = args[1];
			break;
		default:
			fprintf(stderr,
					"usage: %s <pattern> <max errors> [<text file>] [<max occurences print>]\n", args[0]);
			return 1;
	}
	
	// Luetaan tiedosto.
	size_t t_length;
	text = read_file(&t_length, text_file);
	
	// Tarkistetaan hahmon pituus
	size_t p_length = strlen(pattern);
	if (p_length > 63) {
		fprintf(stderr, "Maximum pattern length is 63.\n");
		return 0;
	}
	
	printf("\nSearching pattern '%s' from text with at most %d errors:\n", pattern, k);
	printf("Text contains %d characters.\n", t_length);


	// Rajoitettu dynaaminen ohjelmointi.
	printf("\nUkkonen's restricted dynamic programming:\n");

	reset_clock();							// Aloitetaan ajanotto.
	dp_results = create_int_list();
	dp_find(text, t_length, pattern, strlen(pattern), k, dp_hit);
	double dp_time_used = time_elapsed();	// lopetetaan ajanotto.
	
//	apply_to_int_list(dp_results, print_int);
	printf("Time used: %.2f seconds (100.0 %%)\n", dp_time_used);


	// Myersin bittirinnakkainen algoritmi.
	printf("\nMyers' bit-parallel algorithm:\n");
	
	reset_clock();							// Aloitetaan ajanotto.
	bpm_results = create_int_list();
	bpm_find(text, t_length, pattern, strlen(pattern), k, bpm_hit);
	double bpm_time_used = time_elapsed();	// lopetetaan ajanotto.
	
//	apply_to_int_list(bpm_results, print_int);
	printf("Time used: %.2f seconds (%.1f %%)\n", bpm_time_used, 100.0 * bpm_time_used / dp_time_used);
	
	
	// Bittirinnakkainen NFA-simulointi.
	printf("\nWu-Manber's bit-parallel NFA simulation:\n");
	
	reset_clock();							// Aloitetaan ajanotto.
	bpr_results = create_int_list();
	bpr_find(text, t_length, pattern, strlen(pattern), k, bpr_hit);
	double bpr_time_used = time_elapsed();	// lopetetaan ajanotto.
	
//	apply_to_int_list(bpr_results, print_int);
	printf("Time used: %.2f seconds (%.1f %%)\n\n", bpr_time_used, 100.0 * bpr_time_used / dp_time_used);
	
	
	bool match = equals_int_lists(dp_results, bpr_results) &&
			equals_int_lists(dp_results, bpm_results);
	
	// Tarkastetaan täsmäävätkö algoritmien tulokset.
	printf("Results of all algorithms %smatch.\n",
			match ? "" : "DON'T ");

	// Tulostetaan esiintymien määrä(t).
	if (match) {
		count = 0;
		apply_to_int_list(dp_results, count_node);
		printf("Total count of occurences: %d\n\n", count);
	} else {
		count = 0;
		apply_to_int_list(dp_results, count_node);
		printf("dp result: %d\n\n", count);
		
		count = 0;
		apply_to_int_list(bpm_results, count_node);
		printf("bpm result: %d\n\n", count);
		
		count = 0;
		apply_to_int_list(bpr_results, count_node);
		printf("bpr result: %d\n\n", count);
	}
	
	// Tulostetaan kymmenen ensimmäistä (erillistä) esiintymää.
	if (count > 0 && occ_count > 0) {
		printf("Max first %d distinct occurences:\n", occ_count);
		
		int first_pos = -2;
		int last_pos = -2;
		
		for (IntListNode* n = bpr_results->first; n != NULL; n = n->next) {
		
			if (n->num == last_pos + 1)
				++last_pos;
			else if (last_pos >= 0) {
				printf("%d: ", (last_pos + first_pos) / 2);
				int begin = (last_pos + first_pos) / 2 - p_length - k + 1;
				begin = begin >= 0 ? begin : 0;
				int end = (last_pos + first_pos) / 2 + k;
				end = end < t_length ? end : t_length;
							
				for (char* c = text + begin; c < text + end; ++c)
					if (isprint(*c))
						putchar(*c);
					else if (*c == '\r' || *c == '\n')
						putchar(' ');
					else
						printf("<%02X>", (unsigned char)*c);
				
				printf("\n");
				
				first_pos = n->num;
				last_pos = n->num;			
				--occ_count;

				if (occ_count <= 0)
					break;
			}
			else {
				first_pos = n->num;
				last_pos = n->num;
			}
		}
		
		if (occ_count > 0) {
			printf("%d: ", (last_pos + first_pos) / 2);
			int begin = (last_pos + first_pos) / 2 - p_length - k + 1;
			begin = begin >= 0 ? begin : 0;
			int end = (last_pos + first_pos) / 2 + k;
			end = end < t_length ? end : t_length;
						
			for (char* c = text + begin; c < text + end; ++c)
				if (isprint(*c))
					putchar(*c);
				else if (*c == '\r' || *c == '\n')
					putchar(' ');
				else
					printf("<%02X>", (unsigned char)*c);
			
			printf("\n");
		}
	}
	printf("\n");
	
	delete_int_list(dp_results);
	delete_int_list(bpm_results);
	delete_int_list(bpr_results);
	
	free(text);
	
	return 0;
}
